=== modified file 'find-rcbuggy-problem-packages'
--- find-rcbuggy-problem-packages	2019-09-01 23:35:58 +0000
+++ find-rcbuggy-problem-packages	2019-09-02 01:37:57 +0000
@@ -1,11 +1,43 @@
 #!/usr/bin/python3
 
+# This script highlights packages that are being kept out of Debian
+# testing by release critical bugs whose removal from the Ubuntu devel
+# series would aid the migration of other packages out of proposed.
+#
+# The packages that are being kept out of Debian testing by release
+# critical bugs can be found fairly easily by looking at the output
+# from Debian's britney runs. We do this first (and call these
+# "rc-gone packages").
+#
+# Such packages can inhibit the migration of other packages in two
+# ways:
+#
+# 1) autopkgtest regressions
+# 2) by becoming uninstallable
+#
+# The first is fairly easy to find: scan through Ubuntu's excuses.yaml
+# and look for the rc-gone package in the list of autopkgtest
+# regressions for any package.
+#
+# The second is a bit more mind-bending to detect. If the package is
+# caught up in a transtiton, a rebuild will be attempted as a matter
+# of course, and if this succeeds and passes its autopkgtests then
+# removing the package will not aid proposed migration. So we look for
+# rc-gone packages in proposed that are missing builds or failing
+# autopkgtests. For such source packages, we see if any of their
+# binary packages are reported as being made uninstallable by the
+# migration of any source package in proposed. If so, removing the
+# rc-gone package will help proposed migration.
+
 import argparse
 import glob
 import os
+import shlex
 import sys
 import subprocess
 
+import attr
+
 import yaml
 
 def parse_args():
@@ -18,35 +50,69 @@
 
 args = parse_args()
 
-with open(args.ubuntu_excuses) as fp:
-    ubuntu_excuses = yaml.load(fp, Loader=yaml.CSafeLoader)
-with open(args.ubuntu_update_output) as fp:
+def extract_uninstallable_to_src_pkg(output_fp):
+    # Extract a mapping from binary package name to the set of source
+    # packages whose migration from proposed makes that package
+    # uninstallable.
+
+    # We're looking for sequences of lines like this:
+
+    # skipped: camlp5 (0, 3, 57)
+    #     got: 13+0: a-1:a-0:a-0:i-6:p-0:s-6
+    #      * s390x: hol-light, libaac-tactics-ocaml-dev, libcoq-ocaml-dev, libledit-ocaml-dev, ocaml-ulex08
+
+    # (Britney tries to migrate batches of packages but it always
+    # tries each package on its own as well).
+
+    r = {}
     srcpkg = None
-    uninstallable_to_src_pkg = {}
-    for line in fp:
+    for line in output_fp:
         parts = line.split()
         if len(parts) >= 2:
             if parts[0] == 'skipped:':
                 srcpkg = None
+                # If parts[2] is '(' then this line is about trying to
+                # migrate a single package, which is what we are
+                # looking for.
                 if parts[2].startswith('('):
                     srcpkg = parts[1]
             if srcpkg is not None and parts[0] == '*':
-                for pkg in parts[2:]:
-                    pkg = pkg.strip(',')
-                    uninstallable_to_src_pkg.setdefault(pkg, set()).add(srcpkg)
+                # parts[1] is "${arch}:"
+                # parts[2:] is a comma+space separated list of binary package names.
+                for binpkg in parts[2:]:
+                    binpkg = binpkg.strip(',')
+                    r.setdefault(binpkg, set()).add(srcpkg)
+    return r
+
+with open(args.ubuntu_excuses) as fp:
+    ubuntu_excuses = yaml.load(fp, Loader=yaml.CSafeLoader)
+with open(args.ubuntu_update_output) as fp:
+    uninstallable_to_src_pkg = extract_uninstallable_to_src_pkg(fp)
 with open(args.debian_excuses) as fp:
     debian_excuses = yaml.load(fp, Loader=yaml.CSafeLoader)
 
 dist_files = glob.glob(os.path.join(args.dist_dir, "*_Packages"))
 
-rc_gone = set()
+@attr.s
+class RCGone:
+    source_package_name = attr.ib()
+    bugs = attr.ib()
+    block_by_regression = attr.ib(default=attr.Factory(set))
+    block_by_uninstallability = attr.ib(default=attr.Factory(list))
+
+rc_gones = {}
 
 for source in debian_excuses['sources']:
     if source['old-version'] == '-':
         info = source['policy_info']
-        rc_bug_verdict = info.get('rc-bugs', {}).get('verdict')
-        if rc_bug_verdict == "REJECTED_PERMANENTLY":
-            rc_gone.add(source['item-name'])
+        rc_bugs = info.get('rc-bugs', {})
+        if rc_bugs.get('verdict') == "REJECTED_PERMANENTLY":
+            bugs = []
+            for k in 'shared-bugs', 'unique-source-bugs', 'unique-target-bugs':
+                bugs.extend(rc_bugs[k])
+            rc_gones[source['item-name']] = RCGone(
+                source_package_name=source['item-name'],
+                bugs=bugs)
 
 in_proposed_by_autopkgtest_or_missing_binaries = set()
 
@@ -55,21 +121,35 @@
         in_proposed_by_autopkgtest_or_missing_binaries.add(source['item-name'])
         for package, results in sorted(source['policy_info']['autopkgtest'].items()):
             package = package.split('/')[0]
-            if package not in rc_gone:
+            if package not in rc_gones:
                 continue
             for arch, result in sorted(results.items()):
                 outcome, log, history, wtf1, wtf2 = result
                 if outcome == "REGRESSION":
-                    print(package, "blocks", source['item-name'], "by autopkgtest regression")
+                    rc_gones[package].block_by_regression.add(source['item-name'])
                     break
     if 'missing-builds' in source:
         in_proposed_by_autopkgtest_or_missing_binaries.add(source['item-name'])
 
-for pkg in rc_gone:
+for pkg in rc_gones:
     if pkg not in in_proposed_by_autopkgtest_or_missing_binaries:
         continue
     cp = subprocess.run(["grep-dctrl", "-SwnsPackage", pkg] + dist_files, stdout=subprocess.PIPE, encoding='ascii')
-    bin_pkgs = set(cp.stdout.splitlines())
-    for bin_pkg in sorted(bin_pkgs):
-        if bin_pkg in uninstallable_to_src_pkg:
-            print(pkg, "blocks", sorted(uninstallable_to_src_pkg[bin_pkg]), "by uninstallability of", bin_pkg)
+    for bin_pkg in set(cp.stdout.splitlines()):
+        rc_gones[pkg].block_by_uninstallability.extend(uninstallable_to_src_pkg.get(bin_pkg, []))
+
+for _, rc_gone in sorted(rc_gones.items()):
+    if not rc_gone.block_by_regression and not rc_gone.block_by_uninstallability:
+        continue
+    words = [rc_gone.source_package_name, 'has']
+    comment = "removed from testing (Debian "
+    if len(rc_gone.bugs) == 1:
+        comment += "bug #" + str(rc_gone.bugs[0])
+    else:
+        comment += "bugs " + ', '.join('#' + str(bug) for bug in rc_gone.bugs)
+    comment += ")"
+    if rc_gone.block_by_regression:
+        comment += ', blocks {} by regression'.format(', '.join(sorted(rc_gone.block_by_regression)))
+    if rc_gone.block_by_uninstallability:
+        comment += ', blocks {} by uninstallability'.format(', '.join(sorted(rc_gone.block_by_uninstallability)))
+    print("remove-package", "-m", shlex.quote(comment), rc_gone.source_package_name)

